package net.knarcraft.permissionsigns.command;

import org.bukkit.Bukkit;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.TabCompleter;
import org.bukkit.permissions.Permission;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringJoiner;

/**
 * The tab completer for the create command
 */
public class CreateTabCompleter implements TabCompleter {

    private static List<String> plugins;
    private static Map<String, List<String>> permissions;
    private static List<String> numbers;
    private static List<String> empty;
    private static List<String> name;

    @Override
    public List<String> onTabComplete(@NotNull CommandSender sender, @NotNull Command command, @NotNull String alias,
                                      @NotNull String[] args) {
        //First, display all plugins, filtering on written text
        //Second, if a dot has been input, go to the correct branch,
        //Might want to save the plugin list, but generate permission lists on the fly as needed
        if (plugins == null) {
            loadAvailablePermissions();
            numbers = new ArrayList<>();
            numbers.add(String.valueOf(0));
            empty = new ArrayList<>();
            name = new ArrayList<>();
            name.add("<name>");
        }

        if (args.length > 4) {
            return empty;
        } else if (args.length > 2) {
            return numbers;
        } else if (args.length > 1) {
            return tabCompletePermission(args[1]);
        } else {
            return name;
        }
    }

    /**
     * Gets the tab complete value for the permission typed
     *
     * @param typedNode <p>The full permission node typed by the player</p>
     * @return <p>All known valid auto-complete options</p>
     */
    private List<String> tabCompletePermission(String typedNode) {
        StringBuilder permissionListString = new StringBuilder();
        if (typedNode.contains(",")) {
            String[] permissionList = typedNode.split(",");
            for (int i = 0; i < permissionList.length - 1; i++) {
                permissionListString.append(permissionList[i]);
                permissionListString.append(",");
            }
            typedNode = permissionList[permissionList.length - 1];
        }
        String permissionPrefix = permissionListString.toString();
        List<String> output;
        if (typedNode.contains(".")) {
            List<String> matchingPermissions = permissions.get(typedNode.substring(0, typedNode.lastIndexOf(".")));
            if (matchingPermissions == null) {
                if (permissionPrefix.isEmpty()) {
                    output = new ArrayList<>();
                } else {
                    List<String> onlyPrefix = new ArrayList<>();
                    onlyPrefix.add(permissionPrefix);
                    output = onlyPrefix;
                }
            } else {
                //Filter by the typed text
                output = filterMatching(matchingPermissions, typedNode);
            }
        } else {
            output = plugins;
        }

        //Add previous permissions in the comma-separated lists as a prefix
        if (permissionPrefix.isEmpty()) {
            return output;
        } else {
            List<String> prefixed = new ArrayList<>(output.size());
            for (String matchingCompletion : output) {
                prefixed.add(permissionPrefix + matchingCompletion);
            }
            return prefixed;
        }
    }

    /**
     * Find completable strings which match the text typed by the command's sender
     *
     * @param values    <p>The values to filter</p>
     * @param typedText <p>The text the player has started typing</p>
     * @return <p>The given string values which start with the player's typed text</p>
     */
    private List<String> filterMatching(List<String> values, String typedText) {
        List<String> configValues = new ArrayList<>();
        for (String value : values) {
            if (value.toLowerCase().startsWith(typedText.toLowerCase())) {
                configValues.add(value);
            }
        }
        return configValues;
    }

    /**
     * Loads all permissions available from bukkit plugins
     */
    private static void loadAvailablePermissions() {
        plugins = new ArrayList<>();
        permissions = new HashMap<>();

        for (Permission permission : Bukkit.getPluginManager().getPermissions()) {
            loadPermission(permission.getName());
        }
    }

    /**
     * Loads a given permission into the proper lists and maps
     *
     * @param permissionName <p>The permission to load</p>
     */
    private static void loadPermission(String permissionName) {
        String[] permissionParts = permissionName.split("\\.");
        if (permissionParts.length == 1 && !plugins.contains(permissionParts[0])) {
            plugins.add(permissionParts[0]);
        } else if (permissionParts.length > 1) {
            StringJoiner pathJoiner = new StringJoiner(".");
            for (int j = 0; j < permissionParts.length - 1; j++) {
                pathJoiner.add(permissionParts[j]);
            }
            String path = pathJoiner.toString();
            List<String> permissionList = permissions.computeIfAbsent(path, k -> new ArrayList<>());
            permissionList.add(permissionName);

            loadPermission(path);
        }
    }

}
