package net.knarcraft.permissionsigns.listener;

import net.knarcraft.permissionsigns.PermissionSigns;
import net.knarcraft.permissionsigns.formatting.StringFormatter;
import net.knarcraft.permissionsigns.formatting.TranslatableMessage;
import net.knarcraft.permissionsigns.manager.SignManager;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.block.Sign;
import org.bukkit.block.data.type.WallSign;
import org.bukkit.entity.Player;
import org.bukkit.event.Cancellable;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockExplodeEvent;
import org.bukkit.event.block.BlockPistonExtendEvent;
import org.bukkit.event.block.BlockPistonRetractEvent;
import org.bukkit.event.entity.EntityExplodeEvent;

import java.util.ArrayList;
import java.util.List;

import static net.knarcraft.permissionsigns.PermissionSigns.extensiveSignProtectionEnabled;

/**
 * A listener for relevant block events such as signs being broken
 */
public class BlockListener implements Listener {

    @EventHandler(priority = EventPriority.LOWEST)
    public void onPistonExtend(BlockPistonExtendEvent event) {
        preventDestruction(event.getBlocks(), event);
    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onPistonRetract(BlockPistonRetractEvent event) {
        preventDestruction(event.getBlocks(), event);
    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onBlockExplode(BlockExplodeEvent event) {
        preventDestruction(event.blockList(), event);
    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onEntityExplode(EntityExplodeEvent event) {
        preventDestruction(event.blockList(), event);
    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onBlockBreak(BlockBreakEvent event) {
        Block block = event.getBlock();
        Player player = event.getPlayer();

        if (event.isCancelled()) {
            return;
        }

        //Protect the permission sign itself
        protectBlockIfPermissionSign(event, block, player);

        if (extensiveSignProtectionEnabled()) {
            protectSignsInDirection(event, block, player, BlockFace.UP);
            if (event.isCancelled()) {
                return;
            }
            protectSignsInDirection(event, block, player, BlockFace.DOWN);
        }
    }

    /**
     * Protects signs on falling blocks in the given direction
     *
     * @param event     <p>The triggered event</p>
     * @param block     <p>The broken block</p>
     * @param player    <p>The player breaking the block</p>
     * @param direction <p>The direction to check for affected blocks</p>
     */
    private void protectSignsInDirection(Cancellable event, Block block, Player player, BlockFace direction) {
        Block directionBlock = block.getRelative(direction);
        while ((direction == BlockFace.DOWN && directionBlock.getBlockData().getMaterial() ==
                Material.POINTED_DRIPSTONE) || (direction == BlockFace.UP &&
                isAffectedByGravity(directionBlock.getBlockData().getMaterial()))) {
            protectBlockIfPermissionSign(event, directionBlock, player);
            if (event.isCancelled()) {
                return;
            }
            directionBlock = directionBlock.getRelative(direction);
        }
    }

    /**
     * Prevents an event from destroying permission signs
     *
     * @param blocks <p>The blocks affected by the event</p>
     * @param event  <p>The event to cancel if a permission sign is found</p>
     */
    private void preventDestruction(List<Block> blocks, Cancellable event) {
        //Don't do anything if indirect protection is disabled
        if (!PermissionSigns.indirectProtectionEnabled()) {
            return;
        }
        for (Block block : blocks) {
            protectBlockIfPermissionSign(event, block, null);
            if (event.isCancelled()) {
                return;
            }

            if (extensiveSignProtectionEnabled()) {
                protectSignsInDirection(event, block, null, BlockFace.UP);
                if (event.isCancelled()) {
                    return;
                }
                protectSignsInDirection(event, block, null, BlockFace.DOWN);
            }
        }
    }

    /**
     * Protects the given sign
     *
     * @param event  <p>The triggered event</p>
     * @param block  <p>The block to check if it's a sign</p>
     * @param player <p>The player breaking the block</p>
     */
    private void protectBlockIfPermissionSign(Cancellable event, Block block, Player player) {
        //Protect the permission sign itself
        Material material = block.getBlockData().getMaterial();
        if (Tag.SIGNS.isTagged(material)) {
            checkIfBlockIsPermissionSign(block, player, event);
            if (event.isCancelled()) {
                return;
            }
        }

        lookForAdjacentPermissionSigns(event, block, player);
    }

    /**
     * Looks for permission signs adjacent to the current block
     *
     * @param event  <p>The triggered event</p>
     * @param block  <p>The block to check if it's a sign</p>
     * @param player <p>The player breaking the block</p>
     */
    private void lookForAdjacentPermissionSigns(Cancellable event, Block block, Player player) {
        Block aboveBlock = block.getRelative(BlockFace.UP);
        if (Tag.STANDING_SIGNS.isTagged(aboveBlock.getBlockData().getMaterial())) {
            checkIfBlockIsPermissionSign(aboveBlock, player, event);
            if (event.isCancelled()) {
                return;
            }
        }

        //Check for any wall permission signs that may be broken by breaking the block
        for (BlockFace blockFace : getRelevantBlockFaces()) {
            Block adjacentBlock = block.getRelative(blockFace);
            if (Tag.WALL_SIGNS.isTagged(adjacentBlock.getBlockData().getMaterial()) &&
                    blockFace == ((WallSign) adjacentBlock.getBlockData()).getFacing()) {
                checkIfBlockIsPermissionSign(adjacentBlock, player, event);
                if (event.isCancelled()) {
                    return;
                }
            }
        }
    }

    /**
     * Check if the given block is a permission sign, and perform necessary tasks if it is
     *
     * @param block  <p>The block to check</p>
     * @param player <p>The player that caused the event</p>
     * @param event  <p>The triggered event</p>
     */
    private void checkIfBlockIsPermissionSign(Block block, Player player, Cancellable event) {
        Sign sign = (Sign) block.getState();
        boolean registered = SignManager.getSign(sign.getLocation()) != null;
        if (!registered) {
            return;
        }
        if (player == null) {
            event.setCancelled(true);
            return;
        }
        if (!player.hasPermission("permissionsigns.admin.create")) {
            event.setCancelled(true);
            player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.PERMISSION_SIGN_DESTROY_DENY));
        } else {
            if (!player.isSneaking()) {
                event.setCancelled(true);
                player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.PERMISSION_SIGN_DESTROY_SNEAK));
                return;
            }
            SignManager.removeSign(sign.getLocation());
            player.sendMessage(StringFormatter.getTranslatedInfoMessage(TranslatableMessage.PERMISSION_SIGN_REMOVED));
        }
    }

    /**
     * Gets the relevant block faces for checking around a block
     *
     * @return <p>The relevant block faces for checking around a block</p>
     */
    private List<BlockFace> getRelevantBlockFaces() {
        List<BlockFace> relevantBlockFaces = new ArrayList<>();
        relevantBlockFaces.add(BlockFace.WEST);
        relevantBlockFaces.add(BlockFace.EAST);
        relevantBlockFaces.add(BlockFace.NORTH);
        relevantBlockFaces.add(BlockFace.SOUTH);
        return relevantBlockFaces;
    }

    /**
     * Checks whether the given material is affected by gravity
     *
     * @param material <p>The material to check</p>
     * @return <p>True if the material is affected by gravity</p>
     */
    private boolean isAffectedByGravity(Material material) {
        //TODO: Find a better way of deciding which blocks are affected by gravity
        return Tag.SAND.isTagged(material) || Tag.ANVIL.isTagged(material) || material == Material.POINTED_DRIPSTONE ||
                Tag.STANDING_SIGNS.isTagged(material) || material == Material.DRAGON_EGG || material == Material.GRAVEL ||
                material == Material.BLACK_CONCRETE_POWDER || material == Material.BLUE_CONCRETE_POWDER ||
                material == Material.BROWN_CONCRETE_POWDER || material == Material.CYAN_CONCRETE_POWDER ||
                material == Material.LIME_CONCRETE_POWDER || material == Material.GRAY_CONCRETE_POWDER ||
                material == Material.GREEN_CONCRETE_POWDER || material == Material.LIGHT_BLUE_CONCRETE_POWDER ||
                material == Material.MAGENTA_CONCRETE_POWDER || material == Material.PINK_CONCRETE_POWDER ||
                material == Material.LIGHT_GRAY_CONCRETE_POWDER || material == Material.ORANGE_CONCRETE_POWDER ||
                material == Material.RED_CONCRETE_POWDER || material == Material.PURPLE_CONCRETE_POWDER ||
                material == Material.YELLOW_CONCRETE_POWDER || material == Material.WHITE_CONCRETE_POWDER;
    }

}
