package net.knarcraft.permissionsigns.manager;

import net.knarcraft.permissionsigns.PermissionSigns;
import net.knarcraft.permissionsigns.container.TemporaryPermission;
import net.milkbowl.vault.permission.Permission;
import org.bukkit.Bukkit;
import org.bukkit.OfflinePlayer;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.entity.Player;

import java.io.File;
import java.io.IOException;
import java.util.Queue;
import java.util.Random;
import java.util.UUID;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Level;

/**
 * A manager that performs all Permission tasks
 */
public final class PermissionManager {

    private static Permission permission;
    private static Queue<TemporaryPermission> temporaryPermissions = new PriorityBlockingQueue<>();
    private static final File permissionsFile = new File(PermissionSigns.getInstance().getDataFolder(), "data.yml");

    private PermissionManager() {

    }

    /**
     * Initializes the permission manager
     *
     * @param permission <p>The permission object to use for everything permission-related</p>
     */
    public static void initialize(Permission permission) {
        PermissionManager.permission = permission;
    }

    /**
     * Gets the registered temporary permissions
     *
     * @return <p>The registered temporary permissions</p>
     */
    public static Queue<TemporaryPermission> getTemporaryPermissions() {
        return temporaryPermissions;
    }

    /**
     * Grants a permanent permission to a player
     *
     * @param player         <p>The player to grant the permission to</p>
     * @param permissionNode <p>The permission node to grant to the player</p>
     */
    public static void addPermission(Player player, String permissionNode) {
        //Allow world:permission syntax
        if (permissionNode.contains(":")) {
            grantWorldPermission(player, permissionNode, false);
        } else {
            boolean perWorldPermissions = PermissionSigns.usePerWorldPermissions();
            permission.playerAdd(perWorldPermissions ? player.getWorld().getName() : null, player, permissionNode);
        }
    }

    /**
     * Grants a temporary permission to the given player
     *
     * @param player         <p>The player to add the temporary permission to</p>
     * @param permissionNode <p>The permission node to grant</p>
     * @param worldName      <p>The world to add the permission node to</p>
     */
    public static void grantTemporaryPermission(Player player, String permissionNode, String worldName) {
        //Allow world:permission syntax
        if (permissionNode.contains(":")) {
            grantWorldPermission(player, permissionNode, true);
        } else {
            boolean perWorldPermissions = PermissionSigns.usePerWorldPermissions();
            permission.playerAddTransient(perWorldPermissions ? worldName : null, player, permissionNode);
        }
    }

    /**
     * Grants a permission for the world specified in the permission node
     *
     * @param player         <p>The player to grant the permission node to</p>
     * @param permissionNode <p>The permission node to grant</p>
     * @param temporary      <p>Whether the permission node should be set temporarily</p>
     */
    private static void grantWorldPermission(Player player, String permissionNode, boolean temporary) {
        String[] permissionParts = permissionNode.split(":");
        String world = permissionParts[0];
        if (world.equalsIgnoreCase("all") || world.equalsIgnoreCase("any")) {
            world = null;
        }
        if (temporary) {
            permission.playerAddTransient(world, player, permissionParts[1]);
        } else {
            permission.playerAdd(world, player, permissionParts[1]);
        }
    }

    /**
     * Grants a temporary permission to a player
     *
     * @param player         <p>The player to give the permission to</p>
     * @param permissionNode <p>The temporary permission to grant</p>
     * @param duration       <p>The duration for which the player should keep the given permission</p>
     */
    public static void addTemporaryPermission(Player player, String permissionNode, int duration) {
        String world = player.getWorld().getName();
        grantTemporaryPermission(player, permissionNode, world);
        temporaryPermissions.add(new TemporaryPermission(player, permissionNode, duration, world));
        try {
            saveTemporaryPermissions();
        } catch (IOException e) {
            PermissionSigns.getInstance().getLogger().log(Level.SEVERE, "Unable to save temporary permissions! " +
                    "Players will lose their temporary permissions upon server restart!");
        }
    }

    /**
     * Removes a temporary permission
     *
     * @param offlinePlayer  <p>The player to remove the permission from</p>
     * @param permissionNode <p>The permission node to remove from the player</p>
     */
    public static void removeTemporaryPermission(OfflinePlayer offlinePlayer, String permissionNode) {
        Player player = offlinePlayer.getPlayer();
        if (player != null && player.hasPermission(permissionNode)) {
            permission.playerRemoveTransient(null, player, permissionNode);
        }
    }

    /**
     * Checks whether the given player has the given permission
     *
     * @param player         <p>The player to check</p>
     * @param permissionNode <p>The permission node to check for</p>
     * @param world          <p>The world to check for the permission</p>
     * @return <p>True if the player has the permission</p>
     */
    public static boolean hasPermission(Player player, String permissionNode, String world) {
        return permission.playerHas(world, player, permissionNode);
    }

    /**
     * Saves all temporary permissions to a file
     *
     * @throws IOException <p>If unable to write to the file</p>
     */
    public static void saveTemporaryPermissions() throws IOException {
        YamlConfiguration configuration = YamlConfiguration.loadConfiguration(permissionsFile);
        ConfigurationSection permissionSection = configuration.createSection("permissions");
        Random random = new Random();

        temporaryPermissions.forEach((item) -> {
            //A random long is added to prevent any conflicts
            String key = item.getGrantedPlayer().getUniqueId() + "," + item.getGrantedTime() + "," + random.nextLong();
            permissionSection.set(key + ".permissionNode", item.getPermissionNode());
            permissionSection.set(key + ".duration", item.getGrantedDuration());
            permissionSection.set(key + ".world", item.getWorld());
        });
        configuration.save(permissionsFile);
    }

    /**
     * Loads all saved temporary permissions
     */
    public static void loadTemporaryPermissions() {
        YamlConfiguration configuration = YamlConfiguration.loadConfiguration(permissionsFile);
        ConfigurationSection permissionSection = configuration.getConfigurationSection("permissions");
        temporaryPermissions = new PriorityBlockingQueue<>();
        if (permissionSection == null) {
            PermissionSigns.getInstance().getLogger().log(Level.WARNING, "Permission section not found in data.yml");
            return;
        }

        long currentTime = System.currentTimeMillis();
        for (String key : permissionSection.getKeys(false)) {
            String[] identifierParts = key.split(",");
            OfflinePlayer player = Bukkit.getOfflinePlayer(UUID.fromString(identifierParts[0]));
            String permissionNode = permissionSection.getString(key + ".permissionNode");
            long granted = Long.parseLong(identifierParts[1]);
            int duration = permissionSection.getInt(key + ".duration", 1);
            String world = permissionSection.getString(key + ".world", null);

            //Skip any expired temporary permissions
            if (currentTime > granted + (1000L * duration)) {
                continue;
            }
            temporaryPermissions.add(new TemporaryPermission(player, permissionNode, granted, duration, world));
        }
        try {
            saveTemporaryPermissions();
        } catch (IOException ignored) {
        }
    }

}
